#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <time.h>
#include <io.h>
#include <sys/stat.h>

#include "resource.h"
#include "shared.h"

#define	DBGLOG			1		/// set to 1 to log all of our I/O to BuddyLog, set 0 to ...uh... NOT log it
#define	TRACELEN		128		/// how many opcode addresses to buffer

#define	SERIAL_FILE		1		/// treat SERIAL interface as connected to dumb printer, log to file BuddyPrint.txt.
#define	SERIAL_PLOTTER	1		/// treat SERIAL interface as connected to a plotter. Only very minimal "initialization" implementation, non-functional otherwise, just a sample

/*

NOTE: A Buddy can be written to follow various constraints with regard to the Emulator:

1) It can require that Emulator be run first. If, when Buddy starts up, pMailBox doesn't exist, or pMailBox->hWnd[0]==0, then Buddy terminates.
2) It can require that Buddy be run first. If, when starting up, pMailBox exists AND pMailBox->hWnd[0]!=0, then Buddy terminates.
3) It can auto-terminate when Emualator terminates. When BMSG_GONE is received from Emulator, Buddy terminates itself.
4) At some point, Emulator may be modified to all its configuration to auto-load/start a list of Buddies.
5) A Buddy window may be visible or hidden. Just leave (or omit) the ShowWindow() call in WinMain().

*/



LRESULT CALLBACK MainProc(HWND, UINT, WPARAM, LPARAM);

char		szAppName[] = "Series 80 Emulator Buddy SERIAL";
HANDLE		hInstance;
HWND		hWndMain;
char		ScratBuf[8192];

DWORD		SlotNum;						/// S.C. lost (8=S.C. 3, 15=S.C. 10) for Interrupt purposes
DWORD		CCRADR, IBADR;					/// our Serial Card I/O addresses
#define		PSRADR		CCRADR
#define		OBADR		IBADR
BYTE		CCR, PSR, IB, OB;
BYTE		LastCommand;					/// for reading/writing multiple registers after a COMmand, this is the last COMmand that was written (with an "incrementing FIELD")

BYTE		IOPreg[128];

#define		REG_EOLLEN	0x28
#define		REG_EOL1	0x29
#define		REG_EOL2	0x2A
#define		REG_MODFEAT	0x34
#define		REG_TRCTL	0x39
#define		REG_3A		0x3A

#define		IOPS_WANT		0	/// need to send BMSG_WANT and move on to IOPS_HOOK
#define		IOPS_HOOK		1	/// need to ask for available I/O slots, pick one, hook it, and move on to IOPS_IDLE
#define		IOPS_IDLE		2	/// normal state to hang out in
#define		IOPS_RESET1		3	/// RESET has happened, respond with INT and send first 3 (self-test OK)
#define		IOPS_RESET2		4	/// respond with second 3 (self-test OK)
#define		IOPS_INT		5	/// CPU80 has placed us in an INTERRUPTED state to process commands

int			IOPstate=IOPS_WANT;

/// Fake variables for Fake values for Fake 8250 UART
/// In the case of a real UART, these would be replaced/manipulated as appropriate. See Serial8049.txt code.
BYTE	ModemControlReg;	/// STATUS n,2 (UART's Modem Control Register, register 4)
BYTE	ModemStatusReg;		/// STATUS n,3 (LS 4 bits (3->0) are CTS/DSR/RI/RLSD)
BYTE	LineControlReg;		/// STATUS n,4 (UART's Line Control Register, register 3)
BYTE	MSDivisorLatch;		/// STATUS n,6 (UART's MS Divisor Latch, register 1 with DLAB=1)
BYTE	LSDivisorLatch;		/// STATUS n,7 (UART's LS Divisor Latch, register 0 with DLAB=1)

#if DBGLOG
///*************************************************************************************************************************************************************
/// DBGLOG stuff enables TRACING of CPU80 instructions in bank-swapped ROMs (to know where I/O accesses are coming from) and LOGGING of I/O accesses to BuddyLog
///*************************************************************************************************************************************************************
HANDLE		hLogFile=NULL;
WORD		Trace[TRACELEN];	/// a rolling wrap-around buffer to keep track of bank-switched ROM addresses that access our I/O ports
int			NextTrace=0;		/// the next location in Trace[] to store the address of an I/O port access
BYTE		TraceROM;			/// the ROM# of the ROM which LAST accessed our I/O ports
///******************************************************************************************************
void OpenLogFile() {

	if( hLogFile==NULL ) {
		hLogFile = CreateFile("BuddyLog", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if( hLogFile==INVALID_HANDLE_VALUE ) hLogFile = NULL;
		else SetFilePointer(hLogFile, 0, NULL, FILE_END);	// seek to end of file
	}
}

///********************************************
void Log() {

	int		t, i;
	char	*p;
	DWORD	len;

	if( hLogFile ) {
		sprintf(ScratBuf+strlen(ScratBuf), " ROM:%03o ", TraceROM);

		t = 0;
		i = NextTrace-1;
		p = ScratBuf+strlen(ScratBuf);

		do {
			if( i<0 ) i = TRACELEN-1;
			sprintf(p, " %07o", Trace[i--]);
			p += 8;
		} while( ++t < TRACELEN );

		strcat(ScratBuf, "\r\n");
		WriteFile(hLogFile, ScratBuf, strlen(ScratBuf), &len, NULL);
	}
}

///*******************************************
void CloseLogFile() {

	if( hLogFile!=NULL ) CloseHandle(hLogFile);
	hLogFile = NULL;
}
///*************************************************************************************************************************************************************
/// DBGLOG stuff end
///*************************************************************************************************************************************************************
#endif

#if SERIAL_FILE
///*************************************************************************************************************************************************************
/// SERIAL_FILE stuff enables "dumb printer" (ala 710) and logs all PRINTs to select code to BuddyPrint.txt
/// BuddyPrint.txt file is opened at start of an OUTPUT command and closed after CED byte.
///*************************************************************************************************************************************************************
HANDLE	hPrintFile=NULL;
char	BuddyPrintBuf[256];
int		BuddyPrintLen;

///******************************************************************************************************
void OpenPrintFile() {

	if( hPrintFile==NULL ) {
		hPrintFile = CreateFile("BuddyPrint.txt", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if( hPrintFile==INVALID_HANDLE_VALUE ) hPrintFile = NULL;
		else SetFilePointer(hPrintFile, 0, NULL, FILE_END);	// seek to end of file
	}
	BuddyPrintLen = 0;
}

///********************************************
void BuddyPrint(char c) {

	DWORD	len;

	if( hPrintFile ) {
		if( BuddyPrintLen >= sizeof(BuddyPrintBuf) ) {				/// if unexpectedly long output (buffer full)
			WriteFile(hPrintFile, BuddyPrintBuf, BuddyPrintLen, &len, NULL);	/// dump buffer
			BuddyPrintLen = 0;										/// reset buffer len
		}
		BuddyPrintBuf[BuddyPrintLen++] = c;
	}
}

///********************************************
void BuddyPrintEOL() {

	int		i, j;
	DWORD	len;

	if( hPrintFile==NULL ) OpenPrintFile();
	if( hPrintFile!=NULL ) {
		len = IOPreg[REG_EOLLEN] & 0x3F;			/// length of EOL sequence
		if( !(IOPreg[REG_EOLLEN] & 0x40) && len ) {	/// if EOL_TRANSMIT is not disabled AND EOL_LENGTH is not 0
			for(i=0; i<len; i++) {					/// print the EOL sequence, repeating the last EOL char if the LEN is >7
				j = (i<7) ? i : 6;
				BuddyPrint(IOPreg[REG_EOL1 + j]);
			}
		}
	}
}

///*******************************************
void ClosePrintFile() {

	DWORD	len;

	if( hPrintFile!=NULL ) {
		if( BuddyPrintLen ) WriteFile(hPrintFile, BuddyPrintBuf, BuddyPrintLen, &len, NULL);
		CloseHandle(hLogFile);
	}
	hLogFile = NULL;
	BuddyPrintLen = 0;
}
///*************************************************************************************************************************************************************
/// SERIAL_FILE stuff end
///*************************************************************************************************************************************************************
#endif

#if SERIAL_PLOTTER
///*************************************************************************************************************************************************************
/// SERIAL PLOTTER stuff start
///*************************************************************************************************************************************************************
typedef struct {
	int		state;
	BOOL	outputEnabled;
} PLTR;
PLTR	pltr;
char	cmdbuf[256], datbuf[256];
int		cmdptr, datptr, outptr;

///***************************************************
void BuddyPlotter(char c) {

	if( cmdptr < sizeof(cmdbuf)-1 ) cmdbuf[cmdptr++] = OB;
	if( OB==';' ) {
		if( !strncmp(cmdbuf, "IN;", 3) ) {
			cmdptr = datptr = 0;
			pltr.state = 0;
			pltr.outputEnabled = 0;
		} else if( !strncmp(cmdbuf, "IM36;", 5) ) {
			sprintf(datbuf, "0\r\n");			/// return nonsense value
			datptr = strlen(datbuf);			/// set length of datbuf
			outptr = 0;							/// init "output" ptr to start
		} else if( !strncmp(cmdbuf, "OE;", 3) ) {
			if( datptr ) {
				pltr.outputEnabled = 1;
			}
		} else if( !strncmp(cmdbuf, "OP;", 3) ) {
			sprintf(datbuf, "0,0,100,100\r\n");	/// return nonsense value
			datptr = strlen(datbuf);			/// set length of datbuf
			outptr = 0;							/// init "output" ptr to start
		}
		cmdptr = 0;
	}
}

///*************************************************************************************************************************************************************
/// SERIAL PLOTTER stuff end
///*************************************************************************************************************************************************************
#endif

///******************************************
void GenInt() {

	/// assumes, if we're interrupting CapCPU, that we haven't already got a message pending
	/// and even if we do, we overwrite it, otherwise deadlock looms over our shoulder...
	pMailBox->ArgB2E[MailBoxNum][0] = SlotNum;
	pMailBox->MsgB2E[MailBoxNum] = BMSG_INT;
}

///*******************************************
/// We need to close/unhook everything and reset state
void CloseShop() {

#if DBGLOG
	CloseLogFile();
#endif

#if SERIAL_FILE
	ClosePrintFile();
#endif

	/// release our 'taken' I/O addresses
	if( pMailBox->hWnd[0] && IOPstate>=IOPS_HOOK ) {	/// if Emulator is still there and we're hooked into it
		memset((void*)&pMailBox->ArgB2E[MailBoxNum][0], 0, sizeof(pMailBox->ArgB2E[MailBoxNum]));	/// clear all the possible I/O requests
		*(((WORD*)pMailBox->ArgB2E[MailBoxNum])+0) = CCRADR;	/// add the list of I/O addresses we 'took'
		*(((WORD*)pMailBox->ArgB2E[MailBoxNum])+1) = IBADR;
		pMailBox->MsgB2E[MailBoxNum] = BMSG_IOREL;				/// ask Emulator to release our hooks into those I/O address
		while( pMailBox->hWnd[0] && pMailBox->MsgB2E[MailBoxNum] ) ;	/// as long as Emulator doesn't terminate, wait until message is processed
	}
	IOPstate = IOPS_WANT;			/// Emulator called it quits, so reset everything, in case we're sticking around
}

/// NOTE: To truly implement a SERIAL INTERFACE Buddy, you'll need to peruse the Serial8049.txt reverse-engineered disassembly
/// of the Serial Interface's 8049 ROM, and decide what needs doing, how to do it, etc.

///******************************************************************************************************
/// Handle 8049 0x0AF - 0x0C9 code
void HandleModemControl() {

}

///*************************************************************
BYTE ReadStatus(BYTE lc) {

	BYTE	retval=0xFF;

	switch( lc & 0x0F ) {
	 case 0:
		retval = 2;		/// Serial card ID
		break;
	 case 1:
		retval = IOPreg[0x32];	/// Interrupt Cause
		IOPreg[0x32] = 0;	/// only good for one read, auto-cleared
		break;
	 case 2:
		retval = ModemControlReg;
		break;
	 case 3:
		retval = ModemStatusReg;
		break;
	 case 4:
		retval = LineControlReg;
		break;
	 case 5:
		retval = IOPreg[0x34];
		break;
	 case 6:
		retval = MSDivisorLatch;
		break;
	 case 7:
		retval = LSDivisorLatch;
		break;
	 case 8:
		retval = IOPreg[0x35];
		break;
	 case 9:
		retval = IOPreg[0x39] | (IOPreg[0x04] & 0x80);	/// Transmitter/Receiver Control OR'd with Transmit Enable bit
		break;
	 case 10:				/// Line Status (MS to LS: 0, 0, xmitRegEmpty, breakReceived, framingError, parityError, 0, receivedDataAvailable)
		/// NOTE: in case of a real UART situation, above bits need to be set as appropriate. See 7A4 in Serial8049.txt.
		retval = IOPreg[0x36];
		IOPreg[0x36] = 0;
		break;
	 case 11:
		retval = IOPreg[0x37];	/// Error Termination Cause
		IOPreg[0x37] =0;	/// only good for one read, auto-cleared
		break;
	 default:
		break;
	}
	return retval;
}

///***********************************************************************************
void WriteControl(BYTE lc, BYTE ob) {

	int		i;

	switch( lc & 0x1F ) {
	 case 1:	/// CR1 Interrupt Mask
	 case 2:	/// CR2 Modem Control
	 case 3:	/// CR3 BAUD rate
	 case 4:	/// CR4 Line Characteristics
	 case 5:	/// CR5 Modem Features
	 case 6:	/// CR6 Divisor Latch MSbyte
	 case 7:	/// CR7 Divisor Latch LSbyte
	 case 8:	/// CR8 Error Replacement Character
	 case 9:	/// CR9 Transmitter/Receiver Control
	 case 11:	/// CR11 Input Data Control
	 	/// need to do something for the above, with real hardware...
		break;
	 case 12:	/// CR12 Input termination character #1
	 case 13:	/// CR13 Input termination character #2
	 case 14:	/// CR14 Input termination character #3
	 case 15:	/// CR15 Input termination character #4
	 case 16:	/// CR16 Output EOL sequence LENGTH
	 case 17:	/// CR17 EOL character #1
	 case 18:	/// CR18 EOL character #2
	 case 19:	/// CR19 EOL character #3
	 case 20:	/// CR20 EOL character #4
	 case 21:	/// CR21 EOL character #5
	 case 22:	/// CR22 EOL character #6
	 case 23:	/// CR23 EOL character #7
	 	i = 0x24 + ((lc & 0x1F)-12);
	 	IOPreg[i] = ob;
	 	break;
	 case 25:	/// CR25 LSbyte of termination length of transfer
	 case 26:	/// CR26 MSbyte of termination length of transfer
	 	i = 0x20 + ((lc & 0x1F)-12);
	 	IOPreg[i] = ob;
	 	break;
	 case 27:	/// CR27 Input termination character
	 	IOPreg[0x23] = ob;
	 	break;
	 case 28:	/// CR28 ASSERT byte
	 	ModemControlReg = (ob & 7);
	 	break;
	 case 29:	/// CR29 Service Request byte
		/// SET BREAK
		break;
	 default:
		break;
	}
}

///******************************************************************************************************
/// check OBF, deal with it
void HandleOBF() {

	if( PSR & 0x80 ) {	/// if OBF
		if( CCR & 2 ) {				/// it's a COMmand
///*********
/// IS COM *
///*********
			LastCommand = OB;		/// save for OB arguments
			switch( LastCommand & 0xF0) {	/// branch on the upper nybble (the COMMAND portion)
/// READ STATUS
			 case 0x00:
			 	IB = ReadStatus(LastCommand++);
				PSR |= 1;					/// set IBF bit
				break;
/// INPUT
			 case 0x10:
#if SERIAL_PLOTTER
				if( outptr < datptr && pltr.outputEnabled ) pltr.outputEnabled = 2;	/// the actual sending of bytes happens in the IDLE loop
				else {
					/// interrupt CPU80 with an error (no input)?
				}
#endif // SERIAL_PLOTTER
				break;
/// BURST I/O
			 case 0x20:

				break;
/// INTERRUPT CONTROL
			 case 0x30:
				IOPstate = IOPS_INT;		/// fake that we were interrupted (for someone to do BURST I/O), hang out until INT is strobed
				PSR |= 0x28;				/// set FDPX & PACK
				break;
/// INTERFACE CONTROL
			 case 0x40:
				/// A bunch of the "Interface Control" stuff is HPIB-specific, but I include it here for reference (in case anyone uses this as a template for an HPIB emulator)
				switch( LastCommand & 0x0F ) {
				 case 0:	/// ABORT I/O
					break;
				 case 1:	/// Set REN = 1
					break;
				 case 2:	/// Set REN = 0
					break;
				 case 3:	/// Set ATN = 0
					break;
				 case 4:	/// Perform Parallel Poll
					break;
				 case 5:	/// Send "MY TALK ADDRESS"
					break;
				 case 6:	/// Send "MY LISTEN ADDRESS"
					break;
				 case 7:	/// Send EOL character sequence
#if SERIAL_FILE
					BuddyPrintEOL();
#endif
					break;
				 case 8:	/// BREAK I/O
					break;
				 case 9:	/// RESUME I/O
					break;
				 default:
					break;
				}
				break;
/// READ AUXILLIARY
			 case 0x70:	/// Read Auxilliary

				break;
/// WRITE CONTROL
			 case 0x80:
			 case 0x90:	/// Write Control
				break;
/// OUTPUT
			 case 0xA0:	/// Output
				if( LastCommand & 2 ) {	/// interrupt-driven OUTPUT

				} else {		/// simple OUTPUT
#if SERIAL_FILE
					if( hPrintFile==NULL ) OpenPrintFile();
#endif
#if SERIAL_PLOTTER
					cmdptr = 0;
#endif // SERIAL_PLOTTER
				}
				break;
/// SEND
			 case 0xB0:	/// Send HPIB commands
				break;
/// WRITE AUXILLIARY
			 case 0xE0:	/// Write Auxilliary

				break;
/// EXTENSION
			 case 0xF0:	/// Extension

				break;
/// shouldn't be needed, as we've covered all the possible cases above
			 default:	/// unused
				break;
			}
		} else {	/// not COM, so must be data
///**********
/// NOT COM *
///**********
			switch( LastCommand & 0xF0 ) {
/// READ STATUS
			 case 0x00:
			 	IB = ReadStatus(LastCommand++);
				PSR |= 1;					/// set IBF bit
				break;
/// INPUT
			 case 0x10:
				break;
/// BURST I/O
			 case 0x20:
				break;
/// INTERRUPT CONTROL
			 case 0x30:
				break;
/// INTERFACE CONTROL
			 case 0x40:
				break;
/// READ AUXILLIARY
			 case 0x70:	/// Read Auxilliary

				break;
/// WRITE CONTROL
			 case 0x80:
			 case 0x90:	/// Write Control
			 	WriteControl(LastCommand++, OB);
				break;
/// OUTPUT
			 case 0xA0:	/// Output
				if( LastCommand & 2 ) {	/// interrupt-driven OUTPUT

				} else {		/// simple OUTPUT
#if SERIAL_FILE
					BuddyPrint(OB);
					if( CCR & 4 ) {	/// CED set, that's the end of this OUTPUT
						BuddyPrintEOL();
						ClosePrintFile();
					}
#endif // SERIAL_FILE

#if SERIAL_PLOTTER
					BuddyPlotter(OB);		/// handle the byte output to the plotter
#endif // SERIAL_PLOTTER
				}
				break;
/// SEND
			 case 0xB0:	/// Send HPIB commands
				break;
/// WRITE AUXILLIARY
			 case 0xE0:	/// Write Auxilliary

				break;
/// EXTENSION
			  case 0xF0:	/// Extension

				break;
			 default:
				break;
			}
		}
		PSR &= ~0x80;		/// indicate we've read the OB, so it's now empty
	}
}

///******************************************************************************************************
/// handle ???
void HandleCBthruE0() {

}
///******************************************************************************************************
/// check if transmitter empty, if so, send next byte (0E2-0E4)
void XMITdata() {

	if( CCR & 4 ) {							/// CED (Command End Data) - flush IB
		/// set INPUT buffer len = 0 (if you've buffered input), CPU80 doesn't want any more
	}
	/// if byte in OB, it's the last byte of output


}

///******************************************************************************************************
/// check for pending UART business, deal with it (0E6-0E9)
void HandleUART() {

}

///******************************************************************************************************
/// check if we've interrupted CPU and deal with it (0EB-0FA)
void HandleINTofCPU() {

}

///******************************************************************************************************
int WINAPI WinMain (HINSTANCE hThisInstance, HINSTANCE hPrevInstance, LPSTR lpszArgument, int nCmdShow) {

    MSG			msg;
    WNDCLASS	wc;		// Data structure for the windowclass
    int			n, i;
    DWORD		adr;
    BYTE		b;

    hInstance = hThisInstance;

    // build the windows class structure
    wc.lpfnWndProc = MainProc;				// This function is called by windows
    wc.cbClsExtra = 0;						// No extra bytes after the window class
    wc.cbWndExtra = 0;						// structure or the window instance
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon (hInstance, MAKEINTRESOURCE(IDI_ICON1));
    wc.hCursor = LoadCursor (NULL, IDC_ARROW);
    wc.lpszMenuName = NULL;					// No menu
    wc.lpszClassName = szAppName;
    wc.style = CS_OWNDC | CS_VREDRAW | CS_HREDRAW;
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);

    // Register the window class, and if it fails quit the program
    if( !RegisterClass(&wc) ) return 0;

	hWndMain = CreateWindow(szAppName, szAppName, WS_OVERLAPPEDWINDOW,
		475, 10, 544, 375, NULL, NULL, hInstance, NULL);

    /// setup the share link
    StartShare();
    if( pMailBox==NULL ) {
		MessageBox(NULL, "Unable to open MAILBOX", szAppName, MB_OK);
		return 0;
    }
    /// find and take an open mailbox
    for(n=1; n<8; n++) {
		if( pMailBox->hWnd[n]==0 ) {
			MailBoxNum = n;
			pMailBox->hWnd[MailBoxNum] = hWndMain;		// Emulator is ALWAYS MAILBOX[0]
			break;
		}
    }
    if( n>=8 ) {	// too many Buddies running, bail out (unlikely, but best to check...)
		EndShare(FALSE);
		return 0;
    }

	/// NOTE: This shows the window. If you don't want a window to show on the screen for your Buddy, comment out the following line.
    ShowWindow(hWndMain, nCmdShow);

    /// =========================================
    /// Windows message pump AND MailBox monitor
    /// =========================================
	while( TRUE ) {
		/// Windows messages take priority, to keep the window responsive.

		/// When there's no Windows message, then we check for Emulator messages.

		///======================================
		/// check if Windows message
		///======================================
		/// we first check if we have a Windows message, before calling GetMessage(),
		/// as otherwise, if we call GetMessage() when there's NO message,
		/// it won't come back until there IS a message, and that will
		/// block our communications between Emulator and Buddy.
		while( !PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE) ) {
			///========================================
			/// check for Emulator message
			///========================================

			/// NOTE: In this sample SERIAL INTERFACE emulation, we're sending a BWANT_TRACE so that
			/// we can trace the Series 80 ROM addresses that were executing leading up to each
			/// I/O access to our S.C. 'slot'. That makes debugging and figuring out, what the system
			/// sends to the Serial Card and what it expects back, much easier.
			/// In a real, 'production' buddy, the BWANT_TRACE would not be done, as we have no other
			/// need for trapping all those instructions, which process slows down the emulation noticeably.

			/// NOTE: I/O interface cards are capable of HALTing the Series 80 CPU by setting the MSbit
			/// of the CCR/PSR register. That bit gets mapped by the Translator IC to pull on the HALTEN
			/// line that freezes the Series 80 CPU in its tracks. I don't know whether any of the extant
			/// I/O interface cards make use of that feature, but should YOU want to emulate that in a
			/// Buddy, instead of setting the MSbit in CCR/PSR, you should send a BMSG_HALT with
			/// pMailBox->ArgB2E[MailBoxNum][0]=1 and Emulator will 'hang' in a tight look, just looking
			/// for further messages from that Buddy. You may then send other 'configuration' messages or,
			/// when you're ready, send another BMSG_HALT message with pMailBox->ArgB2E[MailBoxNum][0]=0,
			/// which will release Emulator from its 'hang' state, resuming normal emulation.

			/// NOTE: Similarly, I/O interface cards can INTERRUPT the Series 80 CPU by setting the LSbit of
			/// the CCR/PSR register. Instead of doing that, Buddies must send a BMSG_INT to Emulator with
			/// pMailBox->ArgB2E[MailBoxNum][0] = the Buddy's I/O "slot number" (see BUDDY.TXT documentation
			/// for the BMSG_INT message, and search for references to 'SlotNum' in this sample Buddy source file).

			/// NOTE: Currently, in this sample Serial Buddy, the BUSY bit in PSR is never set. You COULD
			/// set (and clear) it, if you need a delay for some reason, while acquiring data, as that's
			/// part of the IOP/CPU interface and the system code/ROM code checks for busy during many
			/// I/O transactions. But, you also may be able to get away with never having to set it. You'll
			/// have to be the judge of that. But, since Emulator and Buddy tend to wait for BMSG_* messages
			/// to get handled by the other side, the BUSY bit will usually not be necessary.

			/// We take care of IOPS_WANT state right up-front, so Buddy can get his foot in the door (potentially) BEFORE the Emulator is started
			if( IOPstate==IOPS_WANT ) {		/// haven't yet sent WANTS
#if DBGLOG
				OpenLogFile();												/// open BuddyLog file for watching Serial interactions
#endif
				pMailBox->ArgB2E[MailBoxNum][0] = (1<<BWANT_TRACE);			/// set flags for what we WANT
				pMailBox->ArgB2E[MailBoxNum][1] = 060000 | (077777<<16);	/// set START and END addresses for TRACE
				pMailBox->MsgB2E[MailBoxNum] = BMSG_WANT;					/// send the WANT message
				IOPstate = IOPS_HOOK;		/// bump to next state
			}

			if( pMailBox->hWnd[0] ) {	/// if Emulator is running
///===================================================================================
/// before checking for Emulator messages, take care of STATE-specific actions needed.
///===================================================================================
				if( IOPstate >= IOPS_IDLE && (CCR & 0x80) ) IOPstate = IOPS_RESET1;		/// we're being reset

				switch( IOPstate ) {
				 case IOPS_WANT:		/// handled above in order to "get in line" before Emulator (maybe)
					break;
				 case IOPS_HOOK:		/// haven't yet latched into an I/O port
					if( !pMailBox->MsgB2E[MailBoxNum]  ) {		/// if Emulator has taken care of our WANTS message
						pMailBox->MsgB2E[MailBoxNum] = BMSG_IOACTIVE;			/// send msg to find out what I/O addresses are unused
						while( pMailBox->hWnd[0] && pMailBox->MsgB2E[MailBoxNum] ) ;	/// as long as Emulator doesn't terminate, wait until message is processed
						if( pMailBox->hWnd[0] ) {								/// if Emulator is still here
							for(i=7; i>=0; i--) {									/// for each I/O card select code from 10 down to 3 (7 down to 0)
								n = 0120+2*i;											/// n = "I/O byte offset" 0177400+0120 = 0177520 = address of S.C. 3, 2 addresses per S.C. 'slot'
								if( !((pMailBox->BufB2E[MailBoxNum][n>>3]) & (BYTE)(1<<(n&7))) ) break;			/// check the bitmask for that I/O slot (1 BIT per I/O address). If unused, break out
							}
							if( i>=0 ) {											/// if we found an available S.C. slot
								SlotNum = 8+i;											/// set our SlotNum
								CCRADR = 0177520 + 2*i;									/// set our I/O addresses
								IBADR  = CCRADR+1;
								memset((void*)&pMailBox->ArgB2E[MailBoxNum][0], 0, sizeof(pMailBox->ArgB2E[MailBoxNum]));	/// clear all the possible I/O requests
								*(((WORD*)pMailBox->ArgB2E[MailBoxNum])+0) = CCRADR;	/// add the list of I/O addresses we want to 'take'
								*(((WORD*)pMailBox->ArgB2E[MailBoxNum])+1) = IBADR;
								pMailBox->MsgB2E[MailBoxNum] = BMSG_IOTAKE;				/// ask Emulator to hook us into those I/O address
								while( pMailBox->hWnd[0] && pMailBox->MsgB2E[MailBoxNum] ) ;	/// as long as Emulator doesn't terminate, wait until message is processed
								IOPstate = IOPS_IDLE;
							} else {
								MessageBox(NULL, "No available Select Code!", "DEBUG", MB_OK);
								SendMessage(hWndMain, WM_DESTROY, 0, 0);
							}

						} else CloseShop();										/// else reset to "no Emulator" state
					}
					break;
				 case IOPS_RESET1:
					if( !(CCR & 0x80) ) {		/// RESET was released
						memset(IOPreg, 0, sizeof(IOPreg));	/// clear all the registers

						IOPreg[REG_EOLLEN]	= 2;			/// EOL sequence length
						IOPreg[REG_EOL1]	= '\r';			/// EOL sequence CR
						IOPreg[REG_EOL2]	= '\n';			/// ...LF
						IOPreg[REG_MODFEAT]	= 0x30;			/// Modem Features = full auto-handshake
						IOPreg[REG_TRCTL]	= 0x09;			/// Transmit/Receive Control
						IOPreg[REG_3A]		= 0xFF;

						IB = 3;									/// self-test OK
						PSR |= 1;								/// set IBF for CPU80

						IOPstate = IOPS_RESET2;

						GenInt();							/// generate SVCREQUEST (interrupt CPU80)
					}
					break;
				 case IOPS_RESET2:
					if( !(PSR & 1) ) {						/// if first SELF-TEST-OK byte has been read...
						IB = 3;									/// self-test OK, again
						PSR |= (4+1);							/// set PED & IBF for CPU80

						IOPstate = IOPS_IDLE;					/// go to IDLE
					}
					break;
				 default:
		///============================================================================================
		/// This is the main pump for dealing with events, sent, and received data (i.e., the IDLE LOOP
		///============================================================================================
					if( IOPstate != IOPS_INT ) {		/// if Emulator is letting us 'run' in IDLE
#if SERIAL_PLOTTER
						if( !(PSR & 1) && outptr < datptr && pltr.outputEnabled == 2 ) {	/// if not IBF && data to send && we're enabled to send (and CPU80 has asked for it)
							if( CCR & 4 ) {				/// if CPU80 is terminating the INPUT
								outptr = datptr = 0;	/// nothing in plotter's send buffer
								pltr.outputEnabled = 1;	/// plotter output enabled but not started
								PSR = 0x20;				/// FDPX
							} else {
								IB = datbuf[outptr++];		/// send next byte
								PSR |= 1;					/// set IBF
								if( outptr == datptr ) {	/// end of input
									PSR |= 4;				/// set PED
									outptr = datptr = 0;	/// nothing in plotter's send buffer
									pltr.outputEnabled = 1;	/// plotter output enabled but not started
								}
							}
						}
#endif // SERIAL_PLOTTER
						HandleModemControl();			/// Handle 8049 0x0AF - 0x0C9 code
						HandleOBF();					/// check OBF, deal with it
						HandleCBthruE0();				/// handle ???
						XMITdata();						/// check if transmitter empty, if so, send next byte (0E2-0E4)
						HandleUART();					/// check for pending UART business, deal with it (0E6-0E9)
						HandleINTofCPU();				/// check if we've interrupted CPU and deal with it (0EB-0FA)
					} else {
						if( !(CCR & 1) ) HandleOBF();					/// check OBF, deal with it (if not being interrupted)
					}
					break;
				}

			}
///===========================================
/// if Emulator has a message for us
///===========================================
			if( pMailBox->MsgE2B[MailBoxNum] && (IOPstate >= IOPS_IDLE || pMailBox->MsgE2B[MailBoxNum]==BMSG_GONE) ) {
	/// NOTE: The "message clearing" (pMailBox->MsgE2B[MailBoxNum] = 0) happens on a case-by-case basis, as some might want to clear it IMMEDIATELY if processing can be 'lengthy' (relatively speaking)
				switch( pMailBox->MsgE2B[MailBoxNum] ) {
				 case BMSG_RESET:
					pMailBox->MsgE2B[MailBoxNum] = 0;									/// release our mailbox for the Emulator to execute this instruction (before sending the next message)
					break;
				 case BMSG_PWON:		/// Emulator is either switching Model OR starting up
					pMailBox->MsgE2B[MailBoxNum] = 0;									/// release our mailbox for the Emulator to execute this instruction (before sending the next message)
					break;
				 case BMSG_PWOFF:		/// Emulator is either switching Model OR terminating
					CloseShop();
					pMailBox->MsgE2B[MailBoxNum] = 0;									/// release our mailbox for the Emulator to execute this instruction (before sending the next message)
					break;
				 case BMSG_GONE:
	/// NOTE: If you don't want your Buddy to quit when the Emulator quits, comment out the following two lines.
					pMailBox->MsgE2B[MailBoxNum] = 0;						/// Clear the message so we don't repeat this during the WM_DESTROY/PostQuitMessage() process
					SendMessage(hWndMain, WM_DESTROY, 0, 0);
					break;
				 case BMSG_TRACE:
#if DBGLOG
					Trace[NextTrace] = LOWORD(pMailBox->ArgE2B[MailBoxNum][0]);	/// save address of this instruction
				 	TraceROM = *(((BYTE*)pMailBox->ArgE2B[MailBoxNum])+7);		/// save ROM# of this instruction
				 	if( ++NextTrace >= TRACELEN ) NextTrace = 0;				/// wrap index around
#endif
					pMailBox->MsgE2B[MailBoxNum] = 0;									/// release our mailbox for the Emulator to execute this instruction (before sending the next message)
					break;
				 case BMSG_IO:
				 	/// NOTE: since we didn't BMSG_WANT with BWANT_IO, we'll ONLY get this message when it's OUR 'hooked' I/O addresses
	/// NOTE: We don't need to deal with INTRSC (177500), as Emulator takes care of that for us
				 	adr = LOWORD(pMailBox->ArgE2B[MailBoxNum][0]);
					if( HIWORD(pMailBox->ArgE2B[MailBoxNum][0])==0x0000FFFF ) {
///vvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
///**** READ
						if( adr==IBADR ) {
	///**** IB
#if DBGLOG
	sprintf(ScratBuf, "R  IB %07o %2.2X", (int)IBADR, (int)(DWORD)IB);
	Log();
#endif
							pMailBox->ArgE2B[MailBoxNum][1] = IB;	/// value to be 'read'
							PSR &= ~(8+4+1);						/// clear PACK, PED, & IBF
						} else {
	///**** PSR
#if DBGLOG
	sprintf(ScratBuf, "R PSR %07o %2.2X", (int)PSRADR, (int)(DWORD)PSR);
	Log();
#endif
							pMailBox->ArgE2B[MailBoxNum][1] = PSR;	/// value to be 'read'
						}
///^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
					} else {
///vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
///**** WRITE
						if( adr==OBADR ) {
	///**** OB
							OB = (BYTE)HIWORD(pMailBox->ArgE2B[MailBoxNum][0]);
							PSR |= 0x80;	/// set OBF
#if DBGLOG
	sprintf(ScratBuf, "W  OB %07o %2.2X", (int)OBADR, (int)(DWORD)OB);
	Log();
#endif
						} else {
	///**** CCR
							b = CCR;						/// save 'previous state' of CCR
							CCR = (BYTE)HIWORD(pMailBox->ArgE2B[MailBoxNum][0]);	/// set 'new state' of CCR
#if DBGLOG
	sprintf(ScratBuf, "W CCR %07o %2.2X", (int)CCRADR, (int)(DWORD)CCR);
	Log();
#endif
							/// detect RISING EDGE of INT, set PACK
							if( !(b & 1) && (CCR & 1) ) {				/// if INT was LOW but is going HIGH
								if( (CCR & 1) ) {				/// INT (INTerrupt the IOP), we've caught the RISING EDGE - set PACK in PSR
									PSR |=(0x20 + 0x08);		///			set FDPX & PACK
									if( IOPstate!=IOPS_INT ) {	///
										IOPstate = IOPS_INT;	///		put us in INTERRUPTED state
									} else {					///
										IOPstate = IOPS_IDLE;	///		take us out of INTERRUPTED state
									}
								}
							}
							/// detect FALLING EDGE of INT, clear PACK
							if( (b & 0x01) && !(CCR & 0x01) ) {		/// INT was released
								PSR &= ~8;								/// clear PACK
							}
						}
					}
///^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
					pMailBox->ArgE2B[MailBoxNum][0] = 1;							/// indicate we've handled the access
					pMailBox->MsgE2B[MailBoxNum] = 0;							/// release our mailbox for the Emulator to go again
					break;
				 default:
					pMailBox->MsgE2B[MailBoxNum] = 0;	/// Clear the message to release Emulator for future messages
					break;
				}
			}
		}
///======================================
/// if Windows has a message for us
///======================================
		if( GetMessage(&msg, NULL, 0, 0)==TRUE ) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		} else break;
	}
    return 0;
}

///*******************************************************************************
LRESULT CALLBACK MainProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {

	HDC				hdc;
	PAINTSTRUCT		ps;
/*	HFONT			oldfont, smallfont;
	RECT			rect;
	HBRUSH			oldbrush;
	HPEN			oldpen;
	int				i, h;
*/
	switch( message ) {
		case WM_PAINT:
			hdc = BeginPaint(hwnd, &ps);
/*
			SetBkColor(hdc, PALETTERGB(255,255,255));
			SetTextColor(hdc, PALETTERGB(0,0,0));

			oldpen = SelectObject(hdc, GetStockObject(BLACK_PEN));
			oldbrush = SelectObject(hdc, GetStockObject(WHITE_BRUSH));

//	Rectangle(hdc, 0, 0, 3000, 3000);

			rect.left = 10;
			rect.top  = 10;
			rect.right= 3000;
			rect.bottom=3000;

			smallfont = CreateFont(-36*48/100, 0, 0,0, FW_BOLD, FALSE, FALSE, FALSE, ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, FIXED_PITCH + FF_SWISS, NULL);
			oldfont = (HFONT)SelectObject(hdc, smallfont);
			h = 20;

			for(i=0; i<20; i++) {
				rect.top = 10+i*h;
				if( Keys[i]!=0xFF ) {
					sprintf(ScratBuf, "%s", KeyName[(pMailBox->Model<2)?0:1][Keys[i]]);
					DrawTextEx(hdc, ScratBuf, strlen(ScratBuf), &rect, DT_TOP | DT_LEFT, NULL);
				}
			}
			SelectObject(hdc, oldfont);
			SelectObject(hdc, oldbrush);
			SelectObject(hdc, oldpen);
*/
			EndPaint(hwnd, &ps);
			return 0;
		 case WM_DESTROY:
		 	/// close files and release Emulator hooks
		 	CloseShop();
			/// release the shared memory
        	EndShare(TRUE);
        	/// and quit
            PostQuitMessage (0);       // send a WM_QUIT to the message queue
            return 0;
/*		 case WM_MENUCHAR:
			return (MNC_CLOSE<<16);	// to avoid BEEP when holding ALT down and pressing another key
		 case WM_SYSKEYDOWN:
		 case WM_KEYDOWN:
		 	while( pMailBox->hWnd[0] && pMailBox->MsgB2E[MailBoxNum]!=0 ) ;	/// wait for Emulator to deal with previous message from us
		 	if( pMailBox->hWnd[0] ) {	/// if Emulator is still there
				pMailBox->ArgB2E[MailBoxNum][1] = (BYTE)wParam;		/// set the KEYCOD
				pMailBox->ArgB2E[MailBoxNum][0] = 1;				/// set DOWN
		 pMailBox->MsgB2E[MailBoxNum] = BMSG_KEY;				/// send the KEY message
		 	}
			return 0;
		 case WM_SYSKEYUP:
		 case WM_KEYUP:
		 	while( pMailBox->hWnd[0] && pMailBox->MsgB2E[MailBoxNum]!=0 ) ;	/// wait for Emulator to deal with previous message from us
		 	if( pMailBox->hWnd[0] ) {	/// if Emulator is still there
				pMailBox->ArgB2E[MailBoxNum][1] = (BYTE)wParam;		/// set the KEYCOD
				pMailBox->ArgB2E[MailBoxNum][0] = 0;				/// set UP
				pMailBox->MsgB2E[MailBoxNum] = BMSG_KEY;				/// send the KEY message
		 	}
			return 0;
*/
		 default:
		 	break;
    }

    return DefWindowProc(hwnd, message, wParam, lParam);
}
